import re
import sys


def trim(docstring):
    """trim function from PEP-257"""
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def parse_doc_string(doc_string):
    PARAM_OR_RETURNS_REGEX = re.compile(":(?:param|returns)")
    docstring = trim(doc_string)
    long_description = None

    lines = docstring.split("\n", 1)
    short_description = lines[0]

    if len(lines) > 1:
        long_description = lines[1].strip()

        match = PARAM_OR_RETURNS_REGEX.search(long_description)
        if match:
            long_desc_end = match.start()
            long_description = long_description[:long_desc_end].rstrip()

    return short_description, long_description



import cProfile
import datetime
import marshal
import pstats
import random
import string

import boto3
from dynaconf import settings

from awokado.consts import DEFAULT_ACCESS_CONTROL_HEADERS
from awokado.utils import rand_string


class HttpMiddleware:
    def process_request(self, req, resp):
        """Process the request before routing it.

        Args:
            req: Request object that will eventually be
                routed to an on_* responder method.
            resp: Response object that will be routed to
                the on_* responder.
        """
        # Set content type
        resp.content_type = "application/json"

        if not req.headers:
            return

        # Update https headers
        origin = req.headers.get("HTTP_ORIGIN")
        origin2 = req.headers.get("ORIGIN")
        origin = origin2 or origin

        if settings.AWOKADO_DEBUG:
            resp.append_header(name="Access-Control-Allow-Origin", value=origin)
        else:
            if origin and origin in settings.ORIGIN_HOSTS:
                resp.append_header(
                    name="Access-Control-Allow-Origin", value=origin
                )

        resp.set_headers(
            settings.get(
                "AWOKADO_ACCESS_CONTROL_HEADERS", DEFAULT_ACCESS_CONTROL_HEADERS
            )
        )

    def process_resource(self, req, resp, resource, params):
        """Process the request and resource *after* routing.

        Note:
            This method is only called when the request matches
            a route to a resource.

        Args:
            req: Request object that will be passed to the
                routed responder.
            resp: Response object that will be passed to the
                responder.
            resource: Resource object to which the request was
                routed. May be None if no route was found for
                the request.
            params: A dict-like object representing any
                additional params derived from the route's URI
                template fields, that will be passed to the
                resource's responder method as keyword
                arguments.
        """
        if not settings.get("AWOKADO_DEBUG"):
            return

        profiling_enabled = req.get_param_as_bool("profiling")
        if profiling_enabled:
            profile = cProfile.Profile()
            profile.enable()
            req.profile = profile

        debugger_enabled = req.get_param_as_bool("debug")
        if debugger_enabled:
            from pudb.remote import set_trace

            set_trace(term_size=(100, 50))

    def process_response(self, req, resp, resource, req_succeeded):
        """Post-processing of the response (after routing).

        Args:
            req: Request object.
            resp: Response object.
            resource: Resource object to which the request was
                routed. May be None if no route was found
                for the request.
            req_succeeded: True if no exceptions were raised
                while the awokado processed and routed the
                request; otherwise False.
        """
        if not settings.get("AWOKADO_DEBUG"):
            return

        profiling_enabled = req.get_param_as_bool("profiling")
        if profiling_enabled:
            req.profile.disable()
            save_debug_proiling(req.profile)


def save_debug_proiling(profile):
    if settings.AWOKADO_ENABLE_UPLOAD_DEBUG_PROFILING_TO_S3:
        upload_profiling_info_to_s3(profile)
    else:
        save_profiling_info_to_file(profile)


def save_profiling_info_to_file(profile):
    stats = pstats.Stats(profile)
    marshaled_stats = marshal.dumps(stats.stats)

    now = datetime.datetime.now()
    key = (
        f"{now.strftime('%Y-%m-%d')}"
        f"/ {now.strftime('%Y-%m-%dT%H-%M-%S')}"
        f"- {rand_string()}.prof"
    )
    with open(key, "w") as f:
        f.write(marshaled_stats.decode(encoding="utf-8"))


def upload_profiling_info_to_s3(profile):
    stats = pstats.Stats(profile)
    marshaled_stats = marshal.dumps(stats.stats)
    s3client = boto3.client(
        "s3",
        aws_access_key_id=settings.AWOKADO_AWS_S3_DEBUG_PROFILING_ACCESS_KEY,
        aws_secret_access_key=settings.AWOKADO_AWS_S3_DEBUG_PROFILING_SECRET_KEY,
    )

    now = datetime.datetime.now()
    rand_str = "".join(random.choice(string.ascii_lowercase) for _ in range(10))
    key = (
        f"profiling"
        f"/{now.strftime('%Y-%m-%d')}"
        f"/{now.strftime('%Y-%m-%dT%H-%M-%S')}-{rand_str}"
        f".prof"
    )
    kwargs = dict(
        Body=marshaled_stats,
        Bucket=settings.AWOKADO_AWS_S3_DEBUG_PROFILING_BUCKET_NAME,
        Key=key,
    )

    s3client.put_object(**kwargs)



import os
from typing import Union

from jinja2 import Environment, FileSystemLoader

from awokado.filter_parser import OPERATORS_MAPPING


def get_readme(template_abs_path: Union[str, None]) -> Union[str, None]:
    if not template_abs_path:
        return ""
    env = Environment(
        loader=FileSystemLoader(os.path.dirname(template_abs_path)),
        autoescape=True,
    )
    template = env.get_template(os.path.basename(template_abs_path))
    description = template.render(
        filter_ops=enumerate(OPERATORS_MAPPING.keys(), start=1)
    )
    return description



from typing import Optional

import falcon
from dynaconf import settings
from jose import jwt


def set_bearer_header(resp: falcon.response.Response, payload: dict) -> str:
    token = jwt.encode(
        payload, settings.AWOKADO_AUTH_BEARER_SECRET, algorithm="HS256"
    )
    resp.set_header("Authorization", f"Bearer {token}")
    return token


def get_bearer_payload(req: falcon.request.Request) -> Optional[dict]:
    bearer_header = req.get_header("Authorization")

    if not bearer_header or "Bearer " not in bearer_header:
        return None

    bearer = bearer_header.split(" ")

    if len(bearer) != 2:
        return None

    bearer_token = bearer[1]

    payload = jwt.decode(
        bearer_token, settings.AWOKADO_AUTH_BEARER_SECRET, algorithms=["HS256"]
    )

    return payload



import bulky
import json

import falcon
import sqlalchemy as sa
from marshmallow import Schema
from marshmallow import fields
from marshmallow.fields import List
from marshmallow.schema import SchemaMeta
from clavis import Transaction

from awokado.auth import BaseAuth
from awokado.consts import (
    AUDIT_DEBUG,
    UPDATE,
    BULK_CREATE,
    BULK_UPDATE,
    CREATE,
    OP_IN,
    DELETE,
    OP_CONTAINS,
)

from awokado.custom_fields import ToMany, ToOne
from awokado.db import DATABASE_URL, persistent_engine
from awokado.exceptions import (
    BadRequest,
    BadFilter,
    DeleteResourceForbidden,
    MethodNotAllowed,
    RelationNotFound,
    UnsupportedMethod,
)
from awokado.exceptions import Forbidden
from awokado.filter_parser import (
    filter_value_to_python,
    FilterItem,
    OPERATORS_MAPPING,
)
from awokado.utils import (
    get_sort_way,
    empty_response,
    get_read_params,
    ReadContext,
    has_resource_auth,
)


class ResourceMeta(SchemaMeta):
    def __new__(mcs, cls_name=None, superclasses=None, attributes=None):
        new_resource = super(ResourceMeta, mcs).__new__(
            mcs, cls_name, superclasses, attributes
        )
        new_resource.RESOURCES[new_resource.Meta.name] = new_resource
        res_meta = new_resource.Meta

        if res_meta.name == "base_resource":
            return new_resource
        elif res_meta.name == "_resource":
            return new_resource

        if not hasattr(res_meta, "methods") or not res_meta.methods:
            raise Exception(f"{cls_name} must have Meta.methods")
        if not hasattr(res_meta, "name") or not res_meta.name:
            raise Exception(f"{cls_name} must have Meta.name")
        if not hasattr(res_meta, "auth"):
            raise Exception(f"{cls_name} must have Meta.auth")

        if not hasattr(res_meta, "disable_total"):
            res_meta.disable_total = False

        return new_resource


class BaseResource(Schema, metaclass=ResourceMeta):
    RESOURCES = {}

    class Meta:
        name = "base_resource"
        methods = tuple()
        model = None
        auth = BaseAuth
        skip_doc = True
        disable_total = False

    ###########################################################################
    # Marshmallow validation methods
    ###########################################################################

    def validate_create_request(
        self,
        req: falcon.request.Request,
        resp: falcon.response.Response,
        is_bulk=False,
    ):
        methods = self.Meta.methods
        payload = json.load(req.stream)

        if isinstance(payload.get(self.Meta.name), list):
            request_method = BULK_CREATE
            is_bulk = True
        else:
            request_method = CREATE

        if request_method not in methods:
            raise MethodNotAllowed()

        errors = self.validate(payload.get(self.Meta.name), many=is_bulk)

        if errors:
            raise BadRequest(errors)

        req.stream = payload

    def validate_update_request(
        self, req: falcon.request.Request, resp: falcon.response.Response
    ):
        methods = self.Meta.methods
        if UPDATE not in methods and BULK_UPDATE not in methods:
            raise MethodNotAllowed()

        payload = json.load(req.stream)
        errors = self.validate(
            payload.get(self.Meta.name), partial=True, many=True
        )

        if errors:
            raise BadRequest(errors)

        req.stream = payload

    ###########################################################################
    # Falcon methods
    ###########################################################################

    def on_patch(
        self,
        req: falcon.request.Request,
        resp: falcon.response.Response,
        resource_id: int = None,
    ):
        """
        Update
        """

        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, _ = self.auth(session, req, resp)

            self.validate_update_request(req, resp)

            payload = req.stream

            data = payload[self.Meta.name]
            ids = [d.get(self.Meta.model.id.key) for d in data]

            if hasattr(self.Meta, "auth") and self.Meta.auth is not None:
                self.Meta.auth.can_update(session, user_id, ids)

            self.audit_log(
                f"Update: {self.Meta.name}", payload, user_id, AUDIT_DEBUG
            )

            result = self.update(session, payload, user_id, resource_id)

        resp.body = json.dumps(result, default=str)

    def on_post(
        self, req: falcon.request.Request, resp: falcon.response.Response
    ):
        """
        Create
        """

        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)

            self.validate_create_request(req, resp)

            payload = req.stream

            if has_resource_auth(self):
                self.Meta.auth.can_create(session, user_id, skip_exc=False)

            self.audit_log(
                f"Create: {self.Meta.name}", payload, user_id, AUDIT_DEBUG
            )

            result = self.create(session, payload, user_id)

        resp.body = json.dumps(result, default=str)

    def auth(self, *args, **kwargs):
        raise NotImplementedError("auth method not found")

    def audit_log(self, *args, **kwargs):
        raise NotImplementedError("audit_log method not found")

    def on_get(
        self,
        req: falcon.request.Request,
        resp: falcon.response.Response,
        resource_id: int = None,
    ):
        """
        Read

        :param req: falcon.request.Request
        :param resp: falcon.response.Response
        """
        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)
            params = get_read_params(req, self.__class__)

            result = self.read_handler(
                session, user_id, **params, resource_id=resource_id
            )

        resp.body = json.dumps(result, default=str)

    def on_delete(
        self,
        req: falcon.request.Request,
        resp: falcon.response.Response,
        resource_id: int = None,
    ):
        """
        Delete
        """

        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)

            if not resource_id:
                raise DeleteResourceForbidden(
                    details="Bulk deletion is forbidden"
                )

            if DELETE not in self.Meta.methods:
                raise MethodNotAllowed()

            if has_resource_auth(self):
                self.Meta.auth.can_delete(session, user_id, [resource_id])

            result = self.delete(session, user_id, resource_id)

        resp.body = json.dumps(result, default=str)

    ###########################################################################
    # Resource methods
    ###########################################################################

    def update(
        self, session, payload: dict, user_id: int, resource_id: int = None
    ) -> dict:
        # prepare data for update
        data = payload[self.Meta.name]

        result = self.load(data, many=True, partial=True)
        data_to_update = self._to_update(result)

        ids = [_i.get(self.Meta.model.id.key) for _i in data_to_update]

        session.bulk_update_mappings(self.Meta.model, data_to_update)

        op = OPERATORS_MAPPING[OP_IN]
        result = self.read_handler(
            session=session,
            user_id=user_id,
            filters=[FilterItem("id", op[0], op[1], ids)],
        )

        return result

    def create(self, session, payload: dict, user_id: int) -> dict:
        # prepare data to insert
        data = payload[self.Meta.name]

        if isinstance(data, list):
            result = self.bulk_create(session, user_id, data)
            return result

        result = self.load(data)

        data_to_insert = self._to_create(result)

        # insert to DB
        resource_id = session.execute(
            sa.insert(self.Meta.model)
            .values(data_to_insert)
            .returning(self.Meta.model.id)
        ).scalar()

        result = self.read_handler(
            session=session, user_id=user_id, resource_id=resource_id
        )

        return result

    def bulk_create(self, session, user_id: int, data: list) -> dict:
        result = self.load(data, many=True)

        data_to_insert = []
        for item in result:
            data_to_insert.append(self._to_create(item))

        # insert to DB
        resource_ids = bulky.insert(
            session,
            self.Meta.model,
            data_to_insert,
            returning=[self.Meta.model.id]
        )
        ids = [r.id for r in resource_ids]

        op = OPERATORS_MAPPING[OP_IN]
        result = self.read_handler(
            session=session,
            user_id=user_id,
            filters=[FilterItem("id", op[0], op[1], ids)],
        )

        return result

    def delete(self, session, user_id: int, resource_id: int):
        session.execute(
            sa.delete(self.Meta.model).where(self.Meta.model.id == resource_id)
        )
        return {}

    def _to_update(self, data: list) -> list:
        """
        Prepare resource data for SQLAlchemy update query
        """
        to_update_list = []
        for data_line in data:
            to_update = {}
            for fn, v in data_line.items():
                f = self.fields[fn]
                if isinstance(f, ToMany):
                    continue
                model_field = f.metadata.get("model_field")
                if not model_field:
                    continue
                to_update[model_field.key] = v

            to_update_list.append(to_update)

        return to_update_list

    def _to_create(self, data: dict) -> dict:
        """
        Prepare resource data for SQLAlchemy create query
        """
        to_create = {}
        for fn, v in data.items():
            f = self.fields[fn]
            if isinstance(f, ToMany):
                continue
            model_field = f.metadata["model_field"]
            to_create[model_field.key] = v

        return to_create

    def read_handler(
        self,
        session,
        user_id: int,
        include: list = None,
        filters: [FilterItem] = None,
        sort: list = None,
        resource_id: int = None,
        limit: int = None,
        offset: int = None,
    ) -> dict:

        ctx = ReadContext(
            session,
            type(self),
            user_id,
            include,
            filters,
            sort,
            resource_id,
            limit,
            offset,
        )

        self.read__query(ctx)
        self.read__filtering(ctx)
        self.read__sorting(ctx)

        self.read__pagination(ctx)
        self.read__execute_query(session, ctx)

        if not ctx.obj_ids:
            if ctx.is_list:
                return empty_response(self, ctx.is_list)
            else:
                raise BadRequest("Object Not Found")

        self.read__includes(session, ctx)
        serialized_data = self.read__serializing(ctx)
        return serialized_data

    def read__filtering(self, ctx: ReadContext):
        if not ctx.query:
            return

        resource_fields = self.fields
        filters_to_apply = []

        for f in ctx.query:
            resource_field = resource_fields.get(f.field)

            # if not resource_field:
            #     raise BadFilter(
            #         details="Filed <{}> doesn't exist".format(f.field)
            #     )

            model_field = resource_field.metadata.get("model_field")

            if model_field is None:
                raise BadFilter(filter=f.field)

            value = f.wrapper(f.value)
            value = filter_value_to_python(value)

            list_deserialization = isinstance(value, list) and not isinstance(
                resource_field, List
            )

            if value is not None:
                if list_deserialization:
                    value = [resource_field.deserialize(item) for item in value]
                else:
                    value = resource_field.deserialize(value)

            field_expr = getattr(model_field, f.op)(value)
            filters_to_apply.append(field_expr)

        ctx.q = ctx.q.where(sa.and_(*filters_to_apply))

    def read__sorting(self, ctx: ReadContext):
        if ctx.sort:
            # TODO: add support for routes
            # example: (xxx.names.display_name)
            for sort_item in ctx.sort:
                sort_route, sort_way = get_sort_way(sort_item)

                if sort_route in self.fields.keys():
                    r_field = self.fields.get(sort_route)
                    ctx.q = ctx.q.order_by(
                        sort_way(r_field.metadata.get("model_field"))
                    )

    def read__pagination(self, ctx: ReadContext):
        limit = ctx.limit
        offset = ctx.offset
        if limit:
            ctx.q = ctx.q.limit(limit)
        if offset:
            ctx.q = ctx.q.offset(offset)

    def read__query(self, ctx):
        fields_to_select = {}
        join_expressions = []

        for field_name, field in self.fields.items():
            model_field = field.metadata.get("model_field")
            if model_field is None:
                raise Exception(
                    f"{self.Meta.name}.{field_name} field must have "
                    f"'model_field' argument"
                )

            to_join = field.metadata.get("join")
            if to_join:
                join_expressions.append(to_join)

            if isinstance(field, ToMany):
                fields_to_select[field_name] = sa.func.array_remove(
                    sa.func.array_agg(model_field), None
                )
            elif isinstance(field, ToOne):
                fields_to_select[field_name] = model_field
            else:
                fields_to_select[field_name] = model_field

        q = sa.select([clm.label(lbl) for lbl, clm in fields_to_select.items()])

        if join_expressions:
            joins = sa.outerjoin(
                join_expressions[0].left,
                join_expressions[0].right,
                join_expressions[0].onclause,
            )
            for _j in join_expressions[1:]:
                joins = joins.outerjoin(_j.right, _j.onclause)

            q = q.select_from(joins)

        if not ctx.is_list:
            q = q.where(self.Meta.model.id == ctx.resource_id)

        if has_resource_auth(self):
            q = self.Meta.auth.can_read(ctx, q)

        if join_expressions:
            q = q.group_by(self.Meta.model.id)

        ctx.q = q

    def read__includes(self, session, ctx: ReadContext):
        for i in ctx.include or []:
            if "." in i:
                raise BadRequest()

            elif not isinstance(self.fields.get(i), (ToOne, ToMany)):
                raise RelationNotFound(f"Relation <{i}> not found")

            else:
                related_resource_name = self.fields[i].metadata["resource"]
                related_res = self.RESOURCES[related_resource_name]
                related_field = self.fields[i].metadata.get("model_field")
                method_name = f"get_by_{self.Meta.name.lower()}_ids"

                if not hasattr(related_res, method_name):
                    raise BadRequest(
                        f"Relation {related_resource_name} doesn't ready yet. "
                        f"Ask developers to add {method_name} method "
                        f"to {related_resource_name} resource"
                    )

                related_res_obj = related_res()
                related_data = getattr(related_res_obj, method_name)(
                    session, ctx, related_field
                )
                self.__add_related_payload(ctx, related_res, related_data)

    @staticmethod
    def __add_related_payload(
        ctx: ReadContext, related_res, related_data: list
    ):
        if related_res.Meta.name in ctx.related_payload:
            existing_record_ids = [
                rec["id"] for rec in ctx.related_payload[related_res.Meta.name]
            ]
            for rec in related_data:
                if rec["id"] not in existing_record_ids:
                    ctx.related_payload[related_res.Meta.name].append(rec)
        else:
            ctx.related_payload[related_res.Meta.name] = related_data

    def read__serializing(self, ctx: ReadContext):
        response = {self.Meta.name: ctx.parent_payload}

        response.update(ctx.related_payload)

        if ctx.is_list:
            response = {"payload": response, "meta": None}

            if not self.Meta.disable_total:
                response['meta'] = {"total": ctx.total or 0}

        return response

    def read__execute_query(self, session, ctx: ReadContext):
        if not self.Meta.disable_total:
            ctx.q.append_column(sa.func.count().over().label("total"))

        result = session.execute(ctx.q).fetchall()

        serialized_data = self.dump(result, many=True)

        ctx.obj_ids.extend([_i["id"] for _i in serialized_data])
        ctx.parent_payload = serialized_data

        if serialized_data and not self.Meta.disable_total:
            ctx.total = result[0].total



import sqlalchemy as sa
from sqlalchemy.pool import QueuePool

import clavis
from dynaconf import settings

DATABASE_PASSWORD = settings.DATABASE_PASSWORD
DATABASE_HOST = settings.DATABASE_HOST
DATABASE_USER = settings.DATABASE_USER
DATABASE_PORT = settings.DATABASE_PORT
DATABASE_DB = settings.DATABASE_DB

DATABASE_URL = (
    f"postgresql://"
    f"{DATABASE_USER}:{DATABASE_PASSWORD}@"
    f"{DATABASE_HOST}:{DATABASE_PORT}/"
    f"{DATABASE_DB}"
)

clavis.configure(DATABASE_URL)

persistent_engine = sa.create_engine(
    DATABASE_URL,
    encoding="utf-8",
    echo=settings.get("DB_ECHO", False),
    poolclass=QueuePool,
    pool_size=settings.get("DB_CONN_POOL_SIZE", 10),
    max_overflow=settings.get("DB_CONN_MAX_OVERFLOW", 5),
)



from .auth import AuthError
from .auth import BadEmail
from .auth import BadPassword
from .auth import BadPhone
from .auth import BadPhoneVerificationCode
from .auth import DataError
from .auth import IdentificationFailed
from .auth import PasswordMismatch
from .auth import PasswordResetFailed
from .auth import PhoneVerificationFailed
from .auth import UserNotFound
from .bad_request import BadFilter
from .bad_request import BadLimitOffset
from .bad_request import BadRequest
from .bad_request import InvalidBarcode
from .bad_request import MethodNotAllowed
from .bad_request import UnsupportedMethod
from .bad_request import UnsupportedRequestAttr
from .base import BaseApiException
from .forbidden import CreateResourceForbidden
from .forbidden import DeleteResourceForbidden
from .forbidden import Forbidden
from .forbidden import ReadResourceForbidden
from .forbidden import UpdateResourceForbidden
from .not_found import NotFound
from .not_found import RelationNotFound
from .not_found import ResourceNotFound



import re
from collections import namedtuple
from typing import Iterable

from awokado.consts import (
    OP_LTE,
    OP_EQ,
    OP_GTE,
    OP_ILIKE,
    OP_IN,
    OP_EMPTY,
    OP_CONTAINS,
)
from awokado.exceptions.bad_request import BadFilter

OPERATORS_MAPPING = {
    OP_LTE: ("__le__", lambda v: v),
    OP_EQ: ("__eq__", lambda v: v),
    OP_GTE: ("__ge__", lambda v: v),
    OP_ILIKE: (OP_ILIKE, lambda v: f"%{v}%"),
    OP_IN: ("in_", lambda v: v),
    OP_EMPTY: ("is_", lambda v: None),
    OP_CONTAINS: (OP_CONTAINS, lambda v: v),
}
FilterItem = namedtuple("FilterItem", ("field", "op", "wrapper", "value"))


def filter_value_to_python(value):
    """
    Turn the string `value` into a python object.
    >>> filter_value_to_python([1, 2, 3])
    [1, 2, 3]
    >>> filter_value_to_python(123)
    123
    >>> filter_value_to_python('true')
    True
    >>> filter_value_to_python('False')
    False
    >>> filter_value_to_python('null')
    >>> filter_value_to_python('None')
    >>> filter_value_to_python('Ø')
    u'O'
    """
    if isinstance(value, list):
        return value
    if isinstance(value, int):
        return value

    # Simple values
    if value in ["true", "True", True]:
        value = True
    elif value in ["false", "False", False]:
        value = False
    elif value in ("null", "none", "None", None):
        value = None

    return value


def parse_filters(req_params, resource) -> Iterable[FilterItem]:
    result = []
    r = resource()
    filter_expr = r"(?P<field>({res_fields}))" r"\[" r"(?P<op>[a-z]+)" r"\]"

    res_field_names = "|".join(r.fields.keys())
    filter_expr = filter_expr.format(res_fields=res_field_names)

    for p, v in req_params.items():
        re_result = re.match(filter_expr, p)
        if not re_result:
            continue

        field_name = re_result.group("field")
        op_name = re_result.group("op")
        value = v

        if not all([field_name, op_name, value]):
            continue

        op = OPERATORS_MAPPING.get(op_name)

        if not op:
            raise BadFilter(details="Operator {} doesn't exist".format(op_name))

        if op[0] == "in_" or op[0] == OP_CONTAINS:
            if isinstance(value, list):
                value = value
            elif "," in value:
                value = value.split(",")
            else:
                value = [value]

        if op[0] == "is_" and value == "false":
            op = ("isnot", lambda val: None)

        result.append(FilterItem(field_name, op[0], op[1], value))
    return result



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="awokado",
    version="0.3b8",
    description="Fast and flexible API framework based on Falcon and SQLAlchemy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/5783354/awokado",
    author="Dmitry Karnei",
    author_email="5783354@gmail.com",
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ),
    keywords=" ".join(
        sorted(
            {"api", "rest", "wsgi", "falcon", "sqlalchemy", "sqlalchemy-core"}
        )
    ),
    packages=["awokado", "awokado.exceptions", "awokado.documentation"],
    install_requires=(
        "bcrypt",
        "bulky",
        "boto3",
        "dynaconf",
        "falcon",
        "marshmallow>=3.0.0rc3",
        "pyaml",
        "PyJWT",
        "clavis",
        "python-jose",
        "apispec",
        "jinja2",
    ),
    python_requires=">=3.6",
)



# API Methods
CREATE = "create"
READ = "read"
UPDATE = "update"
BULK_UPDATE = "bulk_update"
BULK_CREATE = "bulk_create"
DELETE = "delete"

# Audit logger level
AUDIT_DEBUG = "DEBUG"
AUDIT_INFO = "INFO"
AUDIT_WARNING = "WARNING"

# Filtering Operators
OP_LTE = "lte"
OP_EQ = "eq"
OP_GTE = "gte"
OP_ILIKE = "ilike"
OP_IN = "in"
OP_EMPTY = "empty"
OP_CONTAINS = "contains"

DEFAULT_ACCESS_CONTROL_HEADERS = [
    [
        "Access-Control-Allow-Headers",
        "Content-Type, X-File-Size, X-File-Name, Authorization",
    ],
    ["Access-Control-Allow-Credentials", "true"],
    ["Access-Control-Allow-Methods", "POST, PATCH, GET, OPTIONS, DELETE"],
    ["Access-Control-Max-Age", "3600"],
]
